<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/

namespace ScriptedBrowser;

/**
 * control function for gecko driver
 */
class Control
{
    /**
     * display id
     *
     * @var int
     */
    private $display;

    /**
     * geckodriver host
     *
     * @var string
     */
    private $driverHost;

    /**
     * geckodriver port
     *
     * @var int
     */
    private $driverPort;

    /**
     * geckodriver url
     *
     * @var string $serverURL
     */
    private $driverURL;

    /**
     * session ID
     *
     * @var string
     */
    private $session;

    /**
     * xdo resource
     *
     * @var resource
     */
    private $xdotool = null;

    /**
     * constructor
     *
     * @param string $driverHost hostname of geckodriver
     * @param int $driverPort port of geckodriver
     * @param int $display display id of vnc
     */
    public function __construct($driverHost, $driverPort, $display)
    {
        $this->driverURL = 'http://' . $driverHost . ':' . $driverPort;
        $this->driverHost = $driverHost;
        $this->driverPort = $driverPort;
        $this->session = '';
        $this->display = $display;
    }

    /**
     * click mouse button
     *
     * @param MouseButton $button button to press
     * @return void
     */
    public function click($button)
    {
        fwrite($this->xdotool, 'click ' . $button->value . "\n");
    }

    /**
     * escape input string
     *
     * @param string $input input string
     * @return string escaped string
     */
    private static function escape(string $input): string
    {
        return preg_replace("/\"/", "\\\"", $input);
    }

    /**
     * execute a command (send to geckodriver)
     *
     * @param string $method HTTP method to use
     * @param string $command command to send
     * @param string $payload optional payload to send (JSON)
     * @return mixed server response
     */
    private function execute($method, $command, $payload = '{}')
    {
        $url = $this->driverURL;
        if ($this->session !== '') {
            $url .= '/session/' . $this->session;
        }
        if ($command !== '') {
            $url .= "/$command";
        }
        $ch = curl_init($url);
        if ($method !== 'GET') {
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $payload);
        }
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HTTPHEADER, ['Content-Type: application/json;charset=UTF-8']);
        $resp = curl_exec($ch);
        curl_close($ch);
        $respJson = json_decode($resp, true);
        if (!is_array($respJson)) {
            echo "resp: '$resp'\n";
            return [];
        }
        if (array_key_exists('value', $respJson)) {
            return $respJson['value'];
        }
        print_r($respJson);
        return [];
    }

    /**
     * execute javascript in browser
     *
     * @param string $script script to execute
     * @return mixed result of javacript
     */
    public function executeScript($script)
    {
        $json = json_encode(['script' => $script, 'args'=>[]]);
        return $this->execute('POST', "execute/sync", $json);
    }

    /**
     * switch to full screen
     *
     * @return array{x:int,y:int,width:int,height:int} fullscreen coordinates
     */
    public function fullScreen()
    {
        return $this->execute('POST', 'window/fullscreen');
    }

    /**
     * get bounding rect of element
     *
     * @param string $element elementId
     * @return array{x:int,y:int,width:int,height:int} coordinates of element
     */
    public function getElementRect($element)
    {
        return $this->execute('GET', "element/$element/rect");
    }

    /**
     * get current URL
     *
     * @return string current URL
     */
    public function getURL() : string
    {
        return $this->execute('GET', 'url');
    }

    /**
     * move mouse to absolute position
     *
     * @param int $x X position to move to
     * @param int $y Y position to move to
     * @return void
     */
    public function mousemove($x, $y)
    {
        fwrite($this->xdotool, "mousemove $x $y\n");
    }

    /**
     * move mouse relative
     * @param int $x X position to move to
     * @param int $y Y position to move to
     * @return void
     */
    public function mousemoveRel($x, $y)
    {
        fwrite($this->xdotool, "mousemove_relative $x $y\n");
    }

    /**
     * open geckodriver if not already running
     *
     * @param ?string $session optional session id
     * @return void
     */
    public function openGeckoDriver(?string $session = null)
    {
        // is geckodriver running?
        $socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        $geckosocket = @socket_connect($socket, $this->driverHost, $this->driverPort);
        if ($geckosocket === false) {
            // start geckodriver
            echo "start geckodriver\n";
            exec("DISPLAY=:". $this->display . " nohup geckodriver 2>/dev/null >/tmp/geckodriver &");
            // wait for gecko driver
            while (@socket_connect($socket, $this->driverHost, $this->driverPort) === false) {
                echo "zzZZ\n";
                sleep(1);
            };
            socket_close($socket);
        }

        $openSession = true;
        if ($session !== null) {
            // todo validate session
            $openSession = false;
            $this->session = $session;
        }

        if ($openSession) {
            $sessionData = $this->execute('POST', "session", '{"capabilities" :  {"alwaysMatch": {"moz:firefoxOptions":{"prefs":{"browser.display.os-zoom-behavior": 0,"layout.css.prefers-color-scheme.content-override": 1}}}, "firstMatch": [{"browserName": "firefox"}]}}');
            $this->session = $sessionData['sessionId'];
            echo 'Session: '. $this->session. "\n";
        }
    }

    /**
     * open xdotool for mouse actions
     *
     * @return void
     */
    public function openInput()
    {
        $this->xdotool = popen('DISPLAY=:' . $this->display . ' xdotool -', 'w');
    }

    /**
     * start vnc session if not already running
     *
     * @return void
     */
    public function openVNC()
    {
        $socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        $socketc = @socket_connect($socket, '127.0.0.1', 5900 + $this->display);
        if ($socketc === false) {
            echo "start vnc\n";
            exec('nohup vncserver :' . $this->display . ' >/dev/null 2>&1 &');
            do {
                $socketc = @socket_connect($socket, '127.0.0.1', 5900 + $this->display);
                if ($socketc !== false) {
                    break;
                }
                echo "zzzZZZ\n";
                sleep(1);
            } while ($socket === false);
        }
        @socket_close($socket);
    }

    /**
     * find elements with given css selector
     *
     * @return string[] element ids
     */
    public function selectCSS(string $css)
    {
        return $this->execute(
            'POST',
            'element',
            "{\"using\": \"css selector\", \"value\": \"" . $this->escape($css) . "\"}"
        );
    }

    /**
     * set session id if already existing
     *
     * @param string $session session id
     * @return void
     */
    public function setSession($session)
    {
        $this->session = $session;
    }

    /**
     * set new URL
     *
     * @param string $url new url
     * @return void
     */
    public function setURL($url)
    {
        $this->execute('POST', 'url', "{\"url\": \"$url\"}");
    }

    /**
     * type with given speed
     *
     * @param float $typeSpeed characters/s
     * @param string $text text to type
     * @return void
     */
    public function type($typeSpeed, $text)
    {
        $delay = 1000 / $typeSpeed;
        fwrite($this->xdotool, "type --delay $delay $text\n");
    }
}
